/*
 * ADXF Library
 * 
 *   Copyright (C) 2016, Jonathan Sevy <jsevy@cs.drexel.edu>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.adxf;


import java.util.Stack;
import java.util.Vector;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.DrawFilter;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Paint.Style;
import android.graphics.Path;
import android.graphics.PathMeasure;
import android.graphics.Picture;
import android.graphics.Point;
import android.graphics.PorterDuff;
import android.graphics.PorterDuffXfermode;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Region;
import android.graphics.Typeface;
import android.text.TextUtils;


/**
 * The Android ADXF library provides support for generation of DXF files for use with CAD programs using standard Android "draw" commands. The library provides a special Canvas subclass, DXFCanvas, that is associated with a DXF document and that renders draw commands into DXF syntax. A sequence of Android Canvas drawing method calls will thus create a structured DXF document that, when opened with a standard CAD program, will display the Canvas image as a CAD design that can be modified within the CAD program.
 * <p>
 * This library is released under the <a href="http://choosealicense.com/licenses/mit/">MIT License.
 * <p>
 * Basic Workflow
 * The class DXFCanvas implements the graphics operations defined by the standard Android Canvas class. To create a DXF drawing, you use the standard Android Canvas drawing calls (including drawing, transformations, etc.) on an instance of DXFCanvas. The DXFCanvas will encode these Android drawing commands as DXF objects in an associated DXFDocument, which can then be saved as a DXF file.
 * <p>
 * A DXFCanvas is associated with a DXFDocument. The basic workflow is:
 * <p>
 * <code><pre>
 * // Create a DXF document and get the associated DXFCanvas
 * DXFDocument dxfDocument = new DXFDocument("Example");
 * DXFCanvas dxfCanvas = dxfDocument.getCanvas();
 * 
 * // Do drawing commands as on any other Canvas
 * Paint paint = new Paint();
 * paint.setColor(Color.RED);
 * paint.setTypeface(Typeface.SERIF);
 * paint.setTextSize(38);
 * paint.setStyle(Style.STROKE);
 * dxfCanvas.drawLine(0, 0,100, 100, paint);
 * dxfCanvas.drawRect(120, 180, 400, 400, paint);
 * dxfCanvas.drawText("Some text", 200, 600, paint);
 * paint.setStyle(Style.FILL);
 * dxfCanvas.skew(0.1f, 0.2f);
 * dxfCanvas.rotate(-35);
 * dxfCanvas.drawCircle(300, 300, 100, paint);
 * 
 * // Write the DXF representation in the document to a file; DXF is just text, so the document generates a String as output
 * String dxfText = dxfDocument.toDXFString();
 * String filePath = "path/to/file.dxf";
 * FileWriter fileWriter = new FileWriter(filePath);
 * fileWriter.write(dxfText);
 * fileWriter.flush();
 * fileWriter.close();
 * </pre></code>
 * <p>
 * <h2>Features</h2>
 * The APDF library provides most of the drawing operations available in the standard Canvas class. However, certain methods are unimplemented at present (some due to lack of support in DXF). These unimplemented operations will throw an UnsupportedOperationException if used. Some additional features currently have no or limited support, as indicated below.
 * <h3>Supported operations</h3>
 * <ul>
 *     <li>Drawing operations, including
 * <ul>
 *     <li>drawLine</li>
 *     <li>drawArc</li>
 *     <li>drawCircle</li>
 *     <li>drawOval</li>
 *     <li>drawRect</li>
 *     <li>drawRoundRect</li>
 *     <li>drawPoints</li>
 *     <li>drawColor</li>
 *     <li>drawPaint</li>
 * </ul>
 * </li>
 *     <li>Text operations drawText, drawPosText, drawTextOnPath</li>
 *     <li>Matrix transformation operations including scale, skew, translate, rotate and concat</li>
 *     <li>State save and restore</li>
 * </ul>
 * <h3>Not supported</h3>
 * <ul>
 *     <li>Dotted lines - represented as solid lines</li>
 *     <li>Underlined and strikethrough text - text represented without underline/strikethrough</li>
 *     <li>Image drawing through drawBitmap - currently ignored</li>
 *     <li>Grid and 3D drawing through drawBitmapMesh and drawVertices - currently ignored</li>
 *     <li>Paint xfermodes - Ignored</li>
 *     <li>drawPaint, drawColor, drawRGB - currently ignored</li>
 *     <li>State-save flags and layer information are ignored; save(flags), saveLayer, and saveLayerAlpha all behave the same as plain save()</li>
 *     <li>quickReject() methods - always return false</li>
 *     <li>drawTextRun method - throws UnsupportedOperationException</li>
 *     <li>Clipping, including clipRect, clipRegion, clipPath, etc., methods - throws UnsupportedOperationException</li>
 *     <li>drawPath method - throws UnsupportedOperationException</li>
 *     <li>drawPicture method - throws UnsupportedOperationException</li>
 *     <li>drawPatch method - throws UnsupportedOperationException</li>
 * </ul>
 * <p>
 *
 * @author Jon Sevy, jsevy@cs.drexel.edu
 * 
 */
public class DXFCanvas extends Canvas
{
    
    /**
     * Constants for stroke cap operators
     */
    public static final int DXF_STROKE_CAP_BUTT = 0;
    public static final int DXF_STROKE_CAP_ROUND = 1;
    public static final int DXF_STROKE_CAP_PROJECTING_SQUARE = 2;
    
    /**
     * Constants for stroke join operators
     */
    public static final int DXF_STROKE_JOIN_MITER = 0;
    public static final int DXF_STROKE_JOIN_ROUND = 1;
    public static final int DXF_STROKE_JOIN_BEVEL = 2;
    
    
    /**
     * Current transform Matrix
     */
    private Matrix javaTransformMatrix;
    private Matrix javaToDXFGraphicsMatrix;
    private Matrix graphicsMatrix;
    
    /**
     * Clip region - initially the largest rectangle
     */
    private Region clipRegion = new Region(Integer.MIN_VALUE, Integer.MIN_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);
    
    /**
     * Class to use in graphics state save/restore stack; especially important because the DXF clip region can't be
     * expanded, only reduced...
     * 
     * @author jsevy
     * 
     */
    private class GraphicsState
    {
        Region clipRegion;
        Matrix javaTransformMatrix;
        
        
        
        public GraphicsState(Matrix matrix, Region region)
        {
            this.clipRegion = new Region(region);
            this.javaTransformMatrix = new Matrix(matrix);
        }
    }
    
    private Stack<GraphicsState> graphicsStateStack = new Stack<GraphicsState>();
        
    /**
     * @hide
     */
    protected int mDensity = Bitmap.DENSITY_NONE;
    
    /**
     * Flag for drawTextRun indicating left-to-right run direction.
     * 
     * @hide
     */
    public static final int DIRECTION_LTR = 0;
    
    /**
     * Flag for drawTextRun indicating right-to-left run direction.
     * 
     * @hide
     */
    public static final int DIRECTION_RTL = 1;
    
    // Maximum bitmap size
    private static final int MAXMIMUM_BITMAP_SIZE = Integer.MAX_VALUE;
    
    // the doc to store the DXF text in
    DXFDocument dxfDocument;
    
    
    
    /**
     * Construct a canvas for DXF drawing into the specified DXF document.
     * 
     * @param dxfDocument   Document into which DXF text representation is generated.
     */
    public DXFCanvas(DXFDocument dxfDocument)
    {
        // the doc to store the DXF text in
        this.dxfDocument = dxfDocument;
        
        // initially, we are limited to the page size
        clipRegion = new Region();
        
        // user-defined Java transform - initialize to identity matrix
        javaTransformMatrix = new Matrix();
        
        // transform from Java Space to DXF Space; y axis flipped
        javaToDXFGraphicsMatrix = new Matrix();
        javaToDXFGraphicsMatrix.setScale(1f, -1f);
        
        // start with identity graphics matrix
        graphicsMatrix = new Matrix();
        
        // combine Java and Java2DXF matrices for overall transform
        setMatrix();
        
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException
     */
    public DXFCanvas(Bitmap bitmap)
    {
        throw new UnsupportedOperationException("Can't set a bitmap on a DXF canvas");
    }
    
    
    /**
     * Indicates whether this Canvas uses hardware acceleration - DXFCanvas does not.
     * 
     * @return Always false
     */
    public boolean isHardwareAccelerated()
    {
        return false;
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException
     */
    public void setBitmap(Bitmap bitmap)
    {
        throw new UnsupportedOperationException("Can't set a bitmap on a DXF canvas");
    }
    
    
    
    /**
     * Set the viewport dimensions if this canvas is GL based. If it is not, this method is ignored and no exception is
     * thrown. Since DXFCanvas is not GL-based, does nothing.
     * 
     * @param width
     *            The width of the viewport
     * @param height
     *            The height of the viewport
     * 
     * @hide
     */
    public void setViewport(int width, int height)
    {
        // do nothing
    }
    
    
    
    /**
     * Return true if the device that the current layer draws into is opaque (i.e. does not support per-pixel alpha).
     * DXFCanvas only supports opaque layers, so always returns true.
     * 
     * @return Always returns true
     */
    public boolean isOpaque()
    {
        return true;
    }
    
    
    
    /**
     * Returns the imageable width of the canvas; not relevant here, so set to max integer
     * 
     * @return the imageable width of the current canvas
     */
    public int getWidth()
    {
        return Integer.MAX_VALUE;
    }
    
    
    
    /**
     * Returns the imageable height of the canvas; not relevant here, so set to max integer
     * 
     * @return the imageable height of the current canvas
     */
    public int getHeight()
    {
        return Integer.MAX_VALUE;
    }
    
    
    
    /**
     * <p>
     * Returns the target density of the canvas.
     * </p>
     * 
     * @return Returns the current target density of the canvas, which is used to determine the scaling factor when
     *         drawing a bitmap into it.
     * 
     * @see #setDensity(int)
     * @see Bitmap#getDensity()
     */
    public int getDensity()
    {
        return mDensity;
    }
    
    
    
    /**
     * <p>
     * Specifies the density for this Canvas' backing bitmap. This modifies the target density of the canvas itself, as
     * well as the density of its backing bitmap via {@link Bitmap#setDensity(int) Bitmap.setDensity(int)}.
     * 
     * @param density
     *            The new target density of the canvas, which is used to determine the scaling factor when drawing a
     *            bitmap into it. Use {@link Bitmap#DENSITY_NONE} to disable bitmap scaling.
     * 
     * @see #getDensity()
     * @see Bitmap#setDensity(int)
     */
    public void setDensity(int density)
    {
        mDensity = density;
    }
    
    
    
    /**
     * Not supported; ignored.
     * 
     * @param density
     */
    public void setScreenDensity(int density)
    {
        // does nothing
    }
    
    
    
    /**
     * Returns the maximum allowed width for bitmaps drawn with this canvas. Attempting to draw with a bitmap wider than
     * this value will result in an error.
     * 
     * @see #getMaximumBitmapHeight()
     */
    public int getMaximumBitmapWidth()
    {
        return MAXMIMUM_BITMAP_SIZE;
    }
    
    
    
    /**
     * Returns the maximum allowed height for bitmaps drawn with this canvas. Attempting to draw with a bitmap taller
     * than this value will result in an error.
     * 
     * @see #getMaximumBitmapWidth()
     */
    public int getMaximumBitmapHeight()
    {
        return MAXMIMUM_BITMAP_SIZE;
    }
    
    
    
    /**
     * Saves the current graphics transformation matrix onto a private stack. Subsequent calls to translate,scale,rotate, 
     * skew,concat will all operate as usual, but when the balancing call to restore() is made, those calls will
     * be forgotten, and the settings that existed before the save() will be reinstated.
     * 
     * @return The value to pass to restoreToCount() to balance this save()
     */
    public int save()
    {
        graphicsStateStack.push(new GraphicsState(javaTransformMatrix, clipRegion));
        return graphicsStateStack.size() - 1;
    }
    
    
    
    /*
     * Flag definitions for saving state // restore the current matrix when restore() is called public static final int
     * MATRIX_SAVE_FLAG = 0x01; // restore the current clip when restore() is called public static final int
     * CLIP_SAVE_FLAG = 0x02; // the layer needs to per-pixel alpha public static final int HAS_ALPHA_LAYER_SAVE_FLAG =
     * 0x04; // the layer needs to 8-bits per color component public static final int FULL_COLOR_LAYER_SAVE_FLAG = 0x08;
     * // clip against the layer's bounds public static final int CLIP_TO_LAYER_SAVE_FLAG = 0x10; // restore everything
     * when restore() is called public static final int ALL_SAVE_FLAG = 0x1F;
     */
    
    /**
     * Based on saveFlags, can save the current matrix and clip onto a private stack. Subsequent calls to
     * translate, scale, rotate, skew, concat will all operate as usual, but when the balancing call to
     * restore() is made, those calls will be forgotten, and the settings that existed before the save() will be
     * reinstated.
     * 
     * Currently behaves the same as save() - ignores saveFlags and simply saves the entire state.
     * 
     * @param saveFlags
     *            flag bits that specify which parts of the Canvas state to save/restore - currently ignored
     * @return The value to pass to restoreToCount() to balance this save()
     */
    public int save(int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     * 
     * @param bounds
     *            Ignored
     * @param paint
     *            Ignored
     * @param saveFlags
     *            Ignored
     * @return value to pass to restoreToCount() to balance this save()
     */
    public int saveLayer(RectF bounds, Paint paint, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     */
    public int saveLayer(float left, float top, float right, float bottom, Paint paint, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     * 
     * @param bounds
     *            Ignored
     * @param alpha
     *            Ignored
     * @param saveFlags
     *            Ignored
     * @return value to pass to restoreToCount() to balance this call
     */
    public int saveLayerAlpha(RectF bounds, int alpha, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This behaves the same as save(), and all arguments are ignored.
     */
    public int saveLayerAlpha(float left, float top, float right, float bottom, int alpha, int saveFlags)
    {
        return save();
    }
    
    
    
    /**
     * This call balances a previous call to save(), and is used to remove all modifications to the matrix state
     * since the last save call. It is an error to call restore() more times than save() was called (throws
     * EmptyStackException).
     */
    public void restore()
    {
        // will throw EmptyStackException if nothing to pop
        GraphicsState graphicsState = graphicsStateStack.pop();
        javaTransformMatrix = graphicsState.javaTransformMatrix;
        clipRegion = graphicsState.clipRegion;
        setMatrix();
    }
    
    
    
    /**
     * Returns the number of matrix states on the Canvas' private stack. This will equal # save() calls - #
     * restore() calls.
     */
    public int getSaveCount()
    {
        return graphicsStateStack.size();
    }
    
    
    
    /**
     * Efficient way to pop any calls to save() that happened after the save count reached saveCount. It is an error for
     * saveCount to be less than 1.
     * 
     * Example: int count = canvas.save(); ... // more calls potentially to save() ... canvas.restoreToCount(count); //
     * now the canvas is back in the same state it was before the initial call to save().
     * 
     * @param saveCount
     *            The save level to restore to.
     */
    public void restoreToCount(int saveCount)
    {
        // pop elements off stack until reach the specified saveCount
        // will throw EmptyStackException if not enough to pop
        GraphicsState graphicsState = graphicsStateStack.pop();
        
        while (graphicsStateStack.size() > saveCount)
        {
            graphicsState = graphicsStateStack.pop();
        }
        
        javaTransformMatrix = graphicsState.javaTransformMatrix;
        clipRegion = graphicsState.clipRegion;
        setMatrix();
    }
    
    
    
    /**
     * Create composite matrix from Java transform and Java-to-DXF transform
     */
    private void setMatrix()
    {
        // set graphics matrix
        graphicsMatrix.setConcat(javaToDXFGraphicsMatrix, javaTransformMatrix);
    }
    
    
    
    /**
     * Add a font style if a corresponding one isn't already present.
     * 
     * @param newPaint
     */
    private DXFStyle addFontStyle(Paint paint)
    {
        
        // set font stuff - typeface and size
        Typeface typeface = paint.getTypeface();
        
        if (typeface == null)
            typeface = Typeface.DEFAULT;
        
        DXFStyle style = new DXFStyle(typeface);
        return dxfDocument.addStyle(style);
       
    }
    
    
    /**
     * Preconcat the current matrix with the specified translation
     * 
     * @param dx
     *            The distance to translate in X
     * @param dy
     *            The distance to translate in Y
     */
    public void translate(float dx, float dy)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.translate(dx, dy);
        javaTransformMatrix.preTranslate(dx, dy);
        setMatrix();
    }
    
    
    
    /**
     * Preconcat the current matrix with the specified scale.
     * 
     * @param sx
     *            The amount to scale in X
     * @param sy
     *            The amount to scale in Y
     */
    public void scale(float sx, float sy)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.scale(sx, sy);
        javaTransformMatrix.preScale(sx, sy);
        setMatrix();
    }
    
    
    
    /*
     * Preconcat the current matrix with the specified scale.
     * 
     * @param sx The amount to scale in X
     * 
     * @param sy The amount to scale in Y
     * 
     * @param px The x-coord for the pivot point (unchanged by the scale)
     * 
     * @param py The y-coord for the pivot point (unchanged by the scale)
     */
    /*
     * public final void scale(float sx, float sy, float px, float py) { translate(px, py); scale(sx, sy);
     * translate(-px, -py); }
     */
    
    /**
     * Preconcat the current matrix with the specified rotation.
     * 
     * @param degrees
     *            The amount to rotate, in degrees
     */
    public void rotate(float degrees)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.rotate(degrees);
        javaTransformMatrix.preRotate(degrees);
        setMatrix();
    }
    
    
    
    /*
     * Preconcat the current matrix with the specified rotation.
     * 
     * @param degrees The amount to rotate, in degrees
     * 
     * @param px The x-coord for the pivot point (unchanged by the rotation)
     * 
     * @param py The y-coord for the pivot point (unchanged by the rotation)
     */
    /*
     * public final void rotate(float degrees, float px, float py) { translate(px, py); rotate(degrees); translate(-px,
     * -py); }
     */
    
    /**
     * Preconcat the current matrix with the specified skew.
     * 
     * @param sx
     *            The amount to skew in X
     * @param sy
     *            The amount to skew in Y
     */
    public void skew(float sx, float sy)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.skew(sx, sy);
        javaTransformMatrix.preSkew(sx, sy);
        setMatrix();
    }
    
    
    
    /**
     * Preconcat the current matrix with the specified matrix.
     * 
     * @param matrix
     *            The matrix to preconcatenate with the current matrix
     */
    public void concat(Matrix matrix)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.concat(matrix);
        javaTransformMatrix.preConcat(matrix);
        setMatrix();
    }
    
    
    
    /**
     * Completely replace the current matrix with the specified matrix. If the matrix parameter is null, then the
     * current matrix is reset to identity.
     * 
     * @param matrix
     *            The matrix to replace the current matrix with. If it is null, set the current matrix to identity.
     */
    public void setMatrix(Matrix matrix)
    {
        // pass to superclass so final method getMatrix returns correct result
        super.setMatrix(matrix);
        javaTransformMatrix.set(matrix);
        setMatrix();
    }
    
    
    
    /**
     * Return, in ctm, the current transformation matrix. This does not alter the matrix in the canvas, but just returns
     * a copy of it.
     */
    public void getMatrix(Matrix ctm)
    {
        ctm.set(javaTransformMatrix);
    }
    
    
    
    /**
     * Return a new matrix with a copy of the canvas' current transformation matrix. Since this is declared final in
     * Canvas class, we can't override, but by using the base class setMatrix method our matrix will be returned by the
     * base class getMatrix method.
     */
    /*
     * public final Matrix getMatrix() { return new Matrix(currentMatrix); }
     */
    
    
    
    /**
     * Modify the current clip with the specified rectangle. Only supports INTERSECT as the operation; throws an
     * UnsupportedOperationException if another operation is supplied
     * 
     * @param rect
     *            The rect to intersect with the current clip
     * @param op
     *            How the clip is modified
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(RectF rect, Region.Op op)
    {
        throw new UnsupportedOperationException("Clipping not supported in DXFCanvas");
    }
    
    
    
    /**
     * Modify the current clip with the specified rectangle. 
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param rect
     *            The rectangle to intersect with the current clip.
     * @param op
     *            How the clip is modified
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(Rect rect, Region.Op op)
    {
        throw new UnsupportedOperationException("Clipping not supported in DXFCanvas");
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param rect
     *            The rectangle to intersect with the current clip.
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(RectF rect)
    {
        return clipRect(rect, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param rect
     *            The rectangle to intersect with the current clip.
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(Rect rect)
    {
        return clipRect(rect, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Modify the current clip with the specified rectangle, which is expressed in local coordinates. 
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param left
     *            The left side of the rectangle to intersect with the current clip
     * @param top
     *            The top of the rectangle to intersect with the current clip
     * @param right
     *            The right side of the rectangle to intersect with the current clip
     * @param bottom
     *            The bottom of the rectangle to intersect with the current clip
     * @param op
     *            How the clip is modified
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(float left, float top, float right, float bottom, Region.Op op)
    {
        return clipRect(new RectF(left, top, right, bottom), op);
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param left
     *            The left side of the rectangle to intersect with the current clip
     * @param top
     *            The top of the rectangle to intersect with the current clip
     * @param right
     *            The right side of the rectangle to intersect with the current clip
     * @param bottom
     *            The bottom of the rectangle to intersect with the current clip
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(float left, float top, float right, float bottom)
    {
        return clipRect(new RectF(left, top, right, bottom));
    }
    
    
    
    /**
     * Intersect the current clip with the specified rectangle, which is expressed in local coordinates.
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param left
     *            The left side of the rectangle to intersect with the current clip
     * @param top
     *            The top of the rectangle to intersect with the current clip
     * @param right
     *            The right side of the rectangle to intersect with the current clip
     * @param bottom
     *            The bottom of the rectangle to intersect with the current clip
     * @return true if the resulting clip is non-empty
     */
    public boolean clipRect(int left, int top, int right, int bottom)
    {
        return clipRect(new RectF(left, top, right, bottom));
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param path
     *            The path to operate on the current clip
     * @param op
     *            How the clip is modified
     * @return true if the resulting is non-empty
     */
    public boolean clipPath(Path path, Region.Op op)
    {
        throw new UnsupportedOperationException("Clipping not supported in DXFCanvas");
    }
    
    
    
    /**
     * Intersect the current clip with the specified path. 
     * 
     * Unsupported; throws an UnsupportedOperationException if called
     * 
     * @param path
     *            The path to intersect with the current clip
     * @return true if the resulting is non-empty
     */
    public boolean clipPath(Path path)
    {
        throw new UnsupportedOperationException("Clipping not supported in DXFCanvas");
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param region
     *            The region to operate on the current clip, based on op
     * @param op
     *            How the clip is modified
     * @return true if the resulting is non-empty
     */
    public boolean clipRegion(Region region, Region.Op op)
    {
        throw new UnsupportedOperationException("Clipping not supported in DXFCanvas");
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param region
     *            The region to operate on the current clip, based on op
     * @return true if the resulting is non-empty
     */
    public boolean clipRegion(Region region)
    {
        return clipRegion(region, Region.Op.INTERSECT);
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     */
    public DrawFilter getDrawFilter()
    {
        throw new UnsupportedOperationException("DrawFilter not supported in DXFCanvas");
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     */
    public void setDrawFilter(DrawFilter filter)
    {
        throw new UnsupportedOperationException("DrawFilter not supported in DXFCanvas");
    }
    
    
    
    public enum EdgeType
    {
        BW(0), // !< treat edges by just rounding to nearest pixel boundary
        AA(1); // !< treat edges by rounding-out, since they may be antialiased
        
        EdgeType(int nativeInt)
        {
            this.nativeInt = nativeInt;
        }
        
        /**
         * @hide
         */
        public final int nativeInt;
    }
    
    
    
    /**
     * Not supported, so always returns false.
     * 
     * @param rect
     *            Ignored
     * @param type
     *            Ignored
     * @return Always returns false
     */
    public boolean quickReject(RectF rect, EdgeType type)
    {
        return false;
    }
    
    
    
    /**
     * Not supported, so always returns false.
     * 
     * @param path
     *            Ignored
     * @param type
     *            Ignored
     * @return Always returns false
     */
    public boolean quickReject(Path path, EdgeType type)
    {
        return false;
    }
    
    
    
    /**
     * Not supported, so always returns false.
     * 
     * @param left
     *            Ignored
     * @param top
     *            Ignored
     * @param right
     *            Ignored
     * @param bottom
     *            Ignored
     * @param type
     *            Ignored
     * @return Always returns false
     */
    public boolean quickReject(float left, float top, float right, float bottom, EdgeType type)
    {
        return false;
    }
    
    
    
    /**
     * Retrieve the clip bounds, returning true if they are non-empty.
     * 
     * Since clipping not supported, always returns true.
     * 
     * @param bounds
     *            Return the clip bounds here. If it is null, ignore it but still return true if the current clip is
     *            non-empty.
     * @return true if the current clip is non-empty.
     */
    public boolean getClipBounds(Rect bounds)
    {
        return true;
    }
    
    
    
    /**
     * Retrieve the clip bounds. Can't override the superclass method, so add superclass calls whenever managing the
     * clip.
     * 
     * @return the clip bounds, or [0, 0, 0, 0] if the clip is empty.
     */
    /*
     * public final Rect getClipBounds() { Rect r = new Rect(); getClipBounds(r); return r; }
     */
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified RGB color.
     * 
     * Unsupported; has no effect if called
     * 
     * @param r
     *            red component (0..255) of the color to draw onto the canvas
     * @param g
     *            green component (0..255) of the color to draw onto the canvas
     * @param b
     *            blue component (0..255) of the color to draw onto the canvas
     */
    public void drawRGB(int r, int g, int b)
    {
        int color = Color.rgb(r, g, b);
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setStyle(Style.FILL);
        
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified ARGB color.
     * 
     * Unsupported; has no effect if called
     * 
     * @param a
     *            alpha component (0..255) of the color to draw onto the canvas
     * @param r
     *            red component (0..255) of the color to draw onto the canvas
     * @param g
     *            green component (0..255) of the color to draw onto the canvas
     * @param b
     *            blue component (0..255) of the color to draw onto the canvas
     */
    public void drawARGB(int a, int r, int g, int b)
    {
        int color = Color.argb(a, r, g, b);
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setStyle(Style.FILL);
        
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified color.
     * 
     * Unsupported; has no effect if called
     * 
     * @param color
     *            the color to draw onto the canvas
     */
    public void drawColor(int color)
    {
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setStyle(Style.FILL);
        
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified color and porter-duff xfermode.
     * 
     * Xfermodes are not currently supported, so always opaque overwrite.
     * 
     * Unsupported; has no effect if called
     * 
     * @param color
     *            the color to draw with
     * @param mode
     *            the porter-duff mode to apply to the color - currently ignored
     */
    public void drawColor(int color, PorterDuff.Mode mode)
    {
        Paint paint = new Paint();
        paint.setColor(color);
        paint.setXfermode(new PorterDuffXfermode(mode));
        drawPaint(paint);
    }
    
    
    
    /**
     * Fill the entire page (restricted to the current clip) with the specified paint.
     * 
     * Unsupported; has no effect if called
     * 
     * @param paint
     *            The paint used to draw onto the canvas
     */
    public void drawPaint(Paint paint)
    {
        // do nothing
    }
    
    
    
    /**
     * Draw a series of points. Each point is centered at the coordinate specified by pts[], and its diameter is
     * specified by the paint's stroke width (as transformed by the canvas' CTM).
     * 
     * @param pts
     *            Array of points to draw [x0 y0 x1 y1 x2 y2 ...]
     * @param offset
     *            Number of values to skip before starting to draw.
     * @param count
     *            The number of values to process, after skipping offset of them. Since one point uses two values, the
     *            number of "points" that are drawn is really (count >> 1).
     * @param paint
     *            The paint used to draw the points
     */
    public void drawPoints(float[] pts, int offset, int count, Paint paint)
    {
        // check to make sure we have enough elements in array; just punt if not
        if ((pts.length < offset + count) || (pts.length - offset < 2))
            return;
        
        int i = offset;
        while ((count - i) >= 2)
        {
            drawPoint(pts[i], pts[i + 1], paint);
            i += 2;
        }
    }
    
    
    
    /**
     * Helper for drawPoints() that assumes you want to draw the entire array
     */
    public void drawPoints(float[] pts, Paint paint)
    {
        drawPoints(pts, 0, pts.length, paint);
    }
    
    
    
    /**
     * Helper for drawPoints() for drawing a single point.
     */
    public void drawPoint(float x, float y, Paint paint)
    {
        // transform the point
        float[] transformedCoords = new float[]{x,y};
        graphicsMatrix.mapPoints(transformedCoords);
        DXFPoint point = new DXFPoint(new RealPoint(transformedCoords[0], transformedCoords[1], 0), paint);
        dxfDocument.addEntity(point);
    }
    
    
    
    /**
     * Draw a line segment with the specified start and stop x,y coordinates, using the specified paint.
     * 
     * @param startX
     *            The x-coordinate of the start point of the line
     * @param startY
     *            The y-coordinate of the start point of the line
     * @param paint
     *            The paint used to draw the line
     */
    public void drawLine(float startX, float startY, float stopX, float stopY, Paint paint)
    {
        float[] coords = new float[]{startX, startY, stopX, stopY};
        drawLines(coords, paint);
    }
    
    
    
    /**
     * Draw a series of lines. Each line is taken from 4 consecutive values in the pts array. Thus to draw 1 line, the
     * array must contain at least 4 values. This is logically the same as drawing the array as follows:
     * drawLine(pts[0], pts[1], pts[2], pts[3]) followed by drawLine(pts[4], pts[5], pts[6], pts[7]) and so on.
     * 
     * @param pts
     *            Array of points to draw [x0 y0 x1 y1 x2 y2 ...]
     * @param offset
     *            Number of values in the array to skip before drawing.
     * @param count
     *            The number of values in the array to process, after skipping "offset" of them. Since each line uses 4
     *            values, the number of "lines" that are drawn is really (count >> 2).
     * @param paint
     *            The paint used to draw the points
     */
    public void drawLines(float[] pts, int offset, int count, Paint paint)
    {
        // check to make sure we have enough elements in array; just punt if not
        if ((pts.length < offset + count) || (pts.length - offset < 4))
            return;      
        
        // transform the coordinates using the current graphics transform matrix
        graphicsMatrix.mapPoints(pts);
        
        int i = offset;
        while ((count - i) >= 4)
        {
            DXFLine line = new DXFLine(new RealPoint(pts[i], pts[i + 1], 0), new RealPoint(pts[i + 2], pts[i + 3], 0), paint);
            dxfDocument.addEntity(line);
            i += 4;
        }
    }
    
    
    
    /**
     * Helper method for drawLines for the common case where all points are to be used.
     */
    public void drawLines(float[] pts, Paint paint)
    {
        drawLines(pts, 0, pts.length, paint);
    }
    
    
    
    /**
     * Draw the specified Rect using the specified paint. The rectangle will be filled or framed based on the Style in
     * the paint.
     * 
     * @param rect
     *            The rect to be drawn
     * @param paint
     *            The paint used to draw the rect
     */
    public void drawRect(RectF r, Paint paint)
    {
        drawRect(r.left, r.top, r.right, r.bottom, paint);
    }
    
    
    
    /**
     * Draw the specified Rect using the specified Paint. The rectangle will be filled or framed based on the Style in
     * the paint.
     * 
     * @param r
     *            The rectangle to be drawn.
     * @param paint
     *            The paint used to draw the rectangle
     */
    public void drawRect(Rect r, Paint paint)
    {
        drawRect(r.left, r.top, r.right, r.bottom, paint);
    }
    
    
    
    /**
     * Draw the specified Rect using the specified paint. The rectangle will be filled or framed based on the Style in
     * the paint.
     * 
     * @param left
     *            The left side of the rectangle to be drawn
     * @param top
     *            The top side of the rectangle to be drawn
     * @param right
     *            The right side of the rectangle to be drawn
     * @param bottom
     *            The bottom side of the rectangle to be drawn
     * @param paint
     *            The paint used to draw the rect
     */
    public void drawRect(float left, float top, float right, float bottom, Paint paint)
    {
        // transform the coordinates using the current graphics transform matrix
        float[] pts = {left, top, right, top, right, bottom, left, bottom};
        graphicsMatrix.mapPoints(pts);
        
        // create as lightweight polyline
        Vector<RealPoint> vertices = new Vector<RealPoint>();
        vertices.add(new RealPoint(pts[0], pts[1], 0));
        vertices.add(new RealPoint(pts[2], pts[3], 0));
        vertices.add(new RealPoint(pts[4], pts[5], 0));
        vertices.add(new RealPoint(pts[6], pts[7], 0));
        
        DXFLWPolyline polyline = new DXFLWPolyline(4 /*vertices*/, vertices, true /*closed*/, paint);
        dxfDocument.addEntity(polyline);
    }
    
    
    
    /**
     * Draw the specified oval using the specified paint. The oval will be filled or framed based on the Style in the
     * paint.
     * 
     * @param oval
     *            The rectangle bounds of the oval to be drawn
     */
    public void drawOval(RectF oval, Paint paint)
    {
        if (oval == null)
        {
            throw new NullPointerException();
        }

        drawArc(oval, 0, 360, false, paint);
    }
    
    
    
    /**
     * Draw the specified circle using the specified paint. If radius is <= 0, then nothing will be drawn. The circle
     * will be filled or framed based on the Style in the paint.
     * 
     * @param cx
     *            The x-coordinate of the center of the circle to be drawn
     * @param cy
     *            The y-coordinate of the center of the circle to be drawn
     * @param radius
     *            The radius of the circle to be drawn
     * @param paint
     *            The paint used to draw the circle
     */
    public void drawCircle(float cx, float cy, float radius, Paint paint)
    {
        
    	if (takesCirclesToCircles(javaTransformMatrix))
    	{
    		// do the common case first: a circle is taken to a circle, so represent as a DXFCircle
    		float[] center = {cx, cy};
    		graphicsMatrix.mapPoints(center);
    		radius = graphicsMatrix.mapRadius(radius);
    		
    		DXFCircle circle = new DXFCircle(new RealPoint(center[0], center[1], 0), radius, paint);
            dxfDocument.addEntity(circle);
    	}
    	else
    	{
    		// transform matrix may be taking this circle to an oval; treat as such   		
    		RectF rect = new RectF(cx - radius /*left*/, cy - radius /*top*/, cx + radius /*right*/, cy + radius /*bottom*/);
            drawOval(rect, paint);
    	}
    	
    	
    }
    
    
    
    /**
     * <p>
     * Draw the specified arc, which will be scaled to fit inside the specified oval.
     * </p>
     * 
     * <p>
     * If the start angle is negative or >= 360, the start angle is treated as start angle modulo 360.
     * </p>
     * 
     * <p>
     * If the sweep angle is >= 360, then the oval is drawn completely. Note that this differs slightly from
     * SkPath::arcTo, which treats the sweep angle modulo 360. If the sweep angle is negative, the sweep angle is
     * treated as sweep angle modulo 360
     * </p>
     * 
     * <p>
     * The arc is drawn clockwise. An angle of 0 degrees correspond to the geometric angle of 0 degrees (3 o'clock on a
     * watch.)
     * </p>
     * 
     * @param oval
     *            The bounds of oval used to define the shape and size of the arc
     * @param startAngle
     *            Starting angle (in degrees) where the arc begins
     * @param sweepAngle
     *            Sweep angle (in degrees) measured clockwise
     * @param useCenter
     *            If true, include the center of the oval in the arc, and close it if it is being stroked. This will
     *            draw a wedge
     * @param paint
     *            The paint used to draw the arc
     */
    public void drawArc(RectF oval, float startAngle, float sweepAngle, boolean useCenter, Paint paint)
    {
        
        // get oval parameters
        float xCen = (oval.right + oval.left) / 2;
        float yCen = (oval.bottom + oval.top) / 2;
        
        float a = (oval.right - oval.left) / 2;
        float b = (oval.bottom - oval.top) / 2;
        
        double startParameter = (startAngle * (Math.PI / 180));
        double endParameter = ((startAngle + sweepAngle) * (Math.PI / 180));
        //double sweepAngleRadians = (sweepAngle * (Math.PI / 180));
        
        RealPoint center = new RealPoint(xCen, yCen, 0);
        RealPoint startPoint = new RealPoint(xCen + a*(float)Math.cos(startParameter), yCen + b*(float)Math.sin(startParameter), 0);
        RealPoint endPoint = new RealPoint(xCen + a*(float)Math.cos(endParameter), yCen + b*(float)Math.sin(endParameter), 0);
        //RealPoint startPointRelative = RealPoint.difference(startPoint, center);
        //RealPoint endPointRelative = RealPoint.difference(endPoint, center);
        
        // The common case where the transform doesn't rotate the axes - not using this since the general case covers everything,
        // and this doesn't buy us anything in the ultimate DXF output - still get an ellipse
        /*
        if (preservesAxisOrientations(javaTransformMatrix))
        {
        	RealPoint transformedCenter = RealPoint.mapPoint(graphicsMatrix, center);
    		RealPoint transformedStartPoint = RealPoint.mapPoint(graphicsMatrix, startPoint);
	        RealPoint transformedEndPoint = RealPoint.mapPoint(graphicsMatrix, endPoint);
	        
	        // map semi axes; since transformation is just a scaling plus translation, gives us new a and b
	        float[] xSemiAxis = {a, 0};
	        float[] ySemiAxis = {0, b};
	        graphicsMatrix.mapVectors(xSemiAxis);
	        graphicsMatrix.mapVectors(ySemiAxis);
	        a = xSemiAxis[0];
	        b = ySemiAxis[1];
    		
    		// need to adjust start and end angles based on java-to-dxf axis flip and clockwise vs counterclockwise angles
    		if (sweepAngleRadians < 0)
    		{
    		    startAngleRadians = endAngleRadians;
    		    sweepAngleRadians = -sweepAngleRadians;
    		}
    		
        	
        	DXFEllipse ellipse = new DXFEllipse(transformedCenter, a, b, startParameter, endParameter, paint);
            dxfDocument.addEntity(ellipse);
            
            // if useCenter is true, need to add lines from arc endpoints to center point
	        if (useCenter)
	        {
	            DXFLine startLine = new DXFLine(transformedCenter, transformedStartPoint, paint);
	            DXFLine endLine = new DXFLine(transformedCenter, transformedEndPoint, paint);
	            
	            dxfDocument.addEntity(startLine);
	            dxfDocument.addEntity(endLine);
	        }
        }
        */
        
        // Here goes the complex stuff for the general transformation case involving rotation and skewing of the ellipse	
       
        // Finding the new major axis under a linear transformation is actually a bit tricky: though an
        // ellipse is mapped to an ellipse under a linear (even affine) transformation, the major axis 
        // of the original ellipse is not mapped to the transformed ellipse's major axis by the transformation.
        // To find the new major axis, we do the following: represent points on the ellipse parametrically, as
        //   (x, y) = (a*cos(t), b*sin(t)) (which we'll use as a column vector). 
        // Representing the linear transformation by the matrix 
        //   (c11, c12)
        //   (c21, c22), 
        // the points of the linearly transformed ellipse become
        //   (x, y) = (c11*a*cos(t) + c12*b*sin(t), c21*a*cos(t) + c22*b*sin(t)).
        // Now the points on the major axis are the points farthest from the center (here 
        // at (0, 0) since we're using just the linear part of the affine transfomation). We can
        // then find these point by solving for t that maximizes the (square of) the distance from the center:
        //   D  =  x^2 + y^2  =  (c11*a*cos(t) + c12*b*sin(t))^2  +  (c21*a*cos(t) + c22*b*sin(t))^2
        // i.e., solve for t such that dD/dt = 0.
        //   dD/dt  =  2 * (c11*a*cos(t) + c12*b*sin(t)) * (-c11*a*sin(t) + c12*b*cos(t))
        //           + 2 * (c21*a*cos(t) + c22*b*sin(t)) * (-c21*a*sin(t) + c22*b*cos(t))
        //          =  2 * { [b^2*(c12^2 + c22^2) - a^2*(c11^2 + c21^2)] * sin(t) * cos(t)
        //                      + a*b*(c11*c12 + c21*c22) * (cos(t)^2 - sin(t)^2) }
        // Setting this equal to 0, and dividing by -2 and cos(t), gives
        //   A*tan(t)^2 + B*tan(t) - A  =  0,
        //      where  A  =  a*b*(c11*c12 + c21*c22)
        //      and    B  =  a^2*(c11^2 + c21^2) - b^2*(c12^2 + c22^2)
        // The quadratic formula then gives us our solution:
        //   tan(t)  =  (-B +/- sqrt(B^2 - 4*A*C))/(2*A)  =  (-B +/- sqrt(B^2 + 4*A^2))/(2*A)
        
        // get matrix coefficients
        float[] basisVects = {1, 0, 0, 1};
        graphicsMatrix.mapVectors(basisVects);
        float c11 = basisVects[0];
        float c12 = basisVects[2];
        float c21 = basisVects[1];
        float c22 = basisVects[3];
        
        float A = a*b*(c11*c12 + c21*c22);
        float B = a*a*(c11*c11 + c21*c21) - b*b*(c12*c12 + c22*c22);
        
        double tangent1;
        //double tangent2;
        if (A != 0)
        {
            tangent1 = (-B + Math.sqrt(B*B + 4*A*A))/(2*A);
            //tangent2 = (-B - Math.sqrt(B*B + 4*A*A))/(2*A);
        }
        else
        {
            tangent1 = 0;
            //tangent2 = 0;
        }
        
        double t1 = Math.atan(tangent1);
        double t2 = t1 + Math.PI/2;
        
        double majorAxisParameter;
        double minorAxisParameter;
        double majorAxisLength;
        double minorAxisLength;
        
        // see which is the major axis
        double D1 = ((c11*a*Math.cos(t1) + c12*b*Math.sin(t1))*(c11*a*Math.cos(t1) + c12*b*Math.sin(t1))  +  (c21*a*Math.cos(t1) + c22*b*Math.sin(t1))*(c21*a*Math.cos(t1) + c22*b*Math.sin(t1)));
        double D2 = ((c11*a*Math.cos(t2) + c12*b*Math.sin(t2))*(c11*a*Math.cos(t2) + c12*b*Math.sin(t2))  +  (c21*a*Math.cos(t2) + c22*b*Math.sin(t2))*(c21*a*Math.cos(t2) + c22*b*Math.sin(t2)));
        
        if (D1 >= D2)
        {
        	majorAxisParameter = t1;
        	minorAxisParameter = t2;
        	majorAxisLength = Math.sqrt(D1);
        	minorAxisLength = Math.sqrt(D2);
        }
        else
        {
        	majorAxisParameter = t2;
        	minorAxisParameter = t1;
        	majorAxisLength = Math.sqrt(D2);
        	minorAxisLength = Math.sqrt(D1);
        }
        
        double transformedAxisRatio = minorAxisLength/majorAxisLength;
        
        
        // get the transformed center, major and (relative) minor aces
        RealPoint transformedCenter = RealPoint.mapPoint(graphicsMatrix, center);
        RealPoint transformedMajorAxisRelative = new RealPoint(c11*a*(float)Math.cos(majorAxisParameter) + c12*b*(float)Math.sin(majorAxisParameter), c21*a*(float)Math.cos(majorAxisParameter) + c22*b*(float)Math.sin(majorAxisParameter), 0);
        RealPoint transformedMinorAxisRelative = new RealPoint(c11*a*(float)Math.cos(minorAxisParameter) + c12*b*(float)Math.sin(minorAxisParameter), c21*a*(float)Math.cos(minorAxisParameter) + c22*b*(float)Math.sin(minorAxisParameter), 0);
        
        // Since angles in dxf always go counterclockwise, we want the minor axis which is counterclockwise from the major axis; check the cross product
        // to see if we have the right one, and negate if not
        RealPoint crossProduct = RealPoint.crossProduct(transformedMajorAxisRelative, transformedMinorAxisRelative);
        if (crossProduct.z < 0)
            transformedMinorAxisRelative = RealPoint.scalarProduct(-1, transformedMinorAxisRelative);
        
        // get transformed start and end points for center lines if needed, as well as to find corresponding ellipse parameters
        RealPoint transformedStartPoint = RealPoint.mapPoint(graphicsMatrix, startPoint);
        RealPoint transformedEndPoint = RealPoint.mapPoint(graphicsMatrix, endPoint);
        RealPoint transformedStartPointRelative = RealPoint.difference(transformedStartPoint, transformedCenter);
        RealPoint transformedEndPointRelative = RealPoint.difference(transformedEndPoint, transformedCenter);
        
        // want the arc that goes between the start and end points; but need to supply the start and end parametric t values,
        // and these are relative to the major axis. So we need to find the t values corresponding to the start and end points.
        // To do this, we decompose relative to the major and minor axes and use the basic ellipse equation to find the corresponding t values.
        // Let (u, v) represent the coordinates of a point on the ellipse in the major/minor axis coordinate system. Then
        //    (u, v)  =  (majorAxisLength*cos(t), minorAxisLength*sin(t)),  or
        //    v/u  =  minorAxisLength/majorAxisLength*tan(t), or   t = atan((v*majorAxisLength)/(u*minorAxisLength)), where
        //    u  =  startVector dotproduct majorAxisVector/magnitude(majorAxisVector), and v  =  same with minorAxisVector
        
        double transformedStartParameter;
        double transformedEndParameter;
        
        float uStart = RealPoint.dotProduct(transformedStartPointRelative, transformedMajorAxisRelative)/RealPoint.magnitude(transformedMajorAxisRelative);
        float vStart = RealPoint.dotProduct(transformedStartPointRelative, transformedMinorAxisRelative)/RealPoint.magnitude(transformedMinorAxisRelative);
        if (uStart == 0)
        {
            if (vStart > 0)
                transformedStartParameter = Math.PI/2;
            else
                transformedStartParameter = -Math.PI/2;
        }
        else
        {
            transformedStartParameter = Math.atan((vStart*majorAxisLength)/(uStart*minorAxisLength));
            if (uStart < 0)
            {
                transformedStartParameter += Math.PI;
            }
        }
        
        
        float uEnd = RealPoint.dotProduct(transformedEndPointRelative, transformedMajorAxisRelative)/RealPoint.magnitude(transformedMajorAxisRelative);
        float vEnd = RealPoint.dotProduct(transformedEndPointRelative, transformedMinorAxisRelative)/RealPoint.magnitude(transformedMinorAxisRelative); 
        if (uEnd == 0)
        {
            if (vEnd > 0)
                transformedEndParameter = Math.PI/2;
            else
                transformedEndParameter = -Math.PI/2;
        }
        else
        {
            transformedEndParameter = Math.atan((vEnd*majorAxisLength)/(uEnd*minorAxisLength));
            if (uEnd < 0)
            {
                transformedEndParameter += Math.PI;
            }
        }
        
        // still need to handle the orientation of the start and end parameters; need to supply them in the right
        // order so we get the proper arc. Since angles in Java go clockwise, we'll need to flip the start and end parameters
        // unless the sweep angle's negative or we don't flip the basis vector orientations. We determine the basic vector 
        // orientations using the cross product.
        double dxfStartParameter = transformedStartParameter;
        double dxfEndParameter = transformedEndParameter;
        
        RealPoint basisVectorI = new RealPoint(1, 0, 0);
        RealPoint basisVectorJ = new RealPoint(0, 1, 0);
        RealPoint transformedBasisVectorI = RealPoint.mapVector(graphicsMatrix, basisVectorI);
        RealPoint transformedBasisVectorJ = RealPoint.mapVector(graphicsMatrix, basisVectorJ);
        RealPoint transformedBasisVectorCrossProduct = RealPoint.crossProduct(transformedBasisVectorI, transformedBasisVectorJ);
        float orientationSign = transformedBasisVectorCrossProduct.z;
        
        if (orientationSign * sweepAngle < 0)
        {
            dxfStartParameter = transformedEndParameter;
            dxfEndParameter = transformedStartParameter;
        }
        
        // finally, create an ellipse with the given parameters
        DXFEllipse ellipse = new DXFEllipse(transformedCenter, transformedMajorAxisRelative, transformedAxisRatio, dxfStartParameter, dxfEndParameter, paint);
        dxfDocument.addEntity(ellipse);
    
        // if useCenter is true, need to add lines from arc endpoints to center point
        if (useCenter)
        {
            DXFLine startLine = new DXFLine(transformedCenter, transformedStartPoint, paint);
            DXFLine endLine = new DXFLine(transformedCenter, transformedEndPoint, paint);
            
            dxfDocument.addEntity(startLine);
            dxfDocument.addEntity(endLine);
        }
        
    }
    
    
    
   
    /**
     * Draw the specified round-rect using the specified paint. The roundrect will be filled or framed based on the
     * Style in the paint.
     * 
     * @param rect
     *            The rectangular bounds of the roundRect to be drawn
     * @param rx
     *            The x-radius of the oval used to round the corners
     * @param ry
     *            The y-radius of the oval used to round the corners
     * @param paint
     *            The paint used to draw the roundRect
     */
    public void drawRoundRect(RectF rect, float rx, float ry, Paint paint)
    {
        if (rect == null)
        {
            throw new NullPointerException();
        }
        
        // ick: draw ovals in corners connected by lines, clockwise from top left
        // note that the subroutines will take care of transforming these as needed
        
        // top line and top-right oval segment
        drawLine(rect.left + rx, rect.top, rect.right - rx, rect.top, paint);
        drawArc(new RectF(rect.right - 2 * rx, rect.top, rect.right, rect.top + 2 * ry), 270, 90, false, paint);
        
        // right line and bottom-right oval segment
        drawLine(rect.right, rect.top+ry, rect.right, rect.bottom - ry, paint);
        drawArc(new RectF(rect.right - 2 * rx, rect.bottom - 2 * ry, rect.right, rect.bottom), 0, 90, false, paint);
        
        // bottom line and bottom-left oval segment
        drawLine(rect.right-rx, rect.bottom, rect.left + rx, rect.bottom, paint);
        drawArc(new RectF(rect.left, rect.bottom - 2 * ry, rect.left + 2 * rx, rect.bottom), 90, 90, false, paint);
        
        // left line and top-left arc
        drawLine(rect.left, rect.bottom-ry,rect.left, rect.top + ry, paint);
        drawArc(new RectF(rect.left, rect.top, rect.left + 2 * rx, rect.top + 2 * ry), 180, 90, false, paint);
        
    }
    
    
    
    /**
     * An extension method that supports the drawing of a B-spline. For the common case of a spline that passes through 
     * the first and last control points, set the multiplicity of these equal to the spline degree plus 1.
     * 
     * @param degree        The degree of the spline; must be > 0
     * @param controlPoints The control points that define the shape of the spline, {x1, y1, x2, y2, ..., xn, yn}.
     * @param multiplicities	The multiplicities for the control points; the number of multiplicity values must equal the
     *                      number of control points (i.e., the array size must be half that of the controlPoints array)
     * @param throughEndpoints	If true, the spline will be forced to pass through the endpoints by setting the end
     * 						control point multiplicities to degree + 1
     * @param paint         The Paint object used to draw the spline
     */
    public void drawSpline(int degree, float[] controlPoints, int[] multiplicities, boolean throughEndpoints, Paint paint)
    {
        Vector<SplineControlPoint> transformedControlPoints = new Vector<SplineControlPoint>();
        
        // transform the control points according to the current transform matrix
        for (int i = 0; i < controlPoints.length/2; i++)
        {
            float[] coords = new float[]{controlPoints[2*i], controlPoints[2*i+1]};
            graphicsMatrix.mapPoints(coords);
            SplineControlPoint transformedControlPoint = new SplineControlPoint(coords[0], coords[1], 0, multiplicities[i]);
            transformedControlPoints.add(transformedControlPoint);
        }
        
        DXFSpline spline = new DXFSpline(degree, transformedControlPoints, throughEndpoints, paint);
        dxfDocument.addEntity(spline);
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param path
     *            The path to be drawn
     * @param paint
     *            The paint used to draw the path
     */
    public void drawPath(Path path, Paint paint)
    {
        // paths not supported at this point...
        throw new UnsupportedOperationException("Operation drawPath not currently supported in DXFCanvas");
    }
    
    
    
    /**
     * Not supported; throws UnsupportedOperationException if called.
     * 
     * @param bitmap
     *            The bitmap to draw as an N-patch
     * @param chunks
     *            The patches information (matches the native struct Res_png_9patch)
     * @param dst
     *            The destination rectangle.
     * @param paint
     *            The paint to draw the bitmap with. may be null
     * 
     * @hide
     */
    public void drawPatch(Bitmap bitmap, byte[] chunks, RectF dst, Paint paint)
    {
        throw new UnsupportedOperationException("Operation drawPatch not currently supported in DXFCanvas");
    }
    
    
    
    /**
     * Draw the specified bitmap, with its top/left corner at (x,y), using the specified paint, transformed by the
     * current matrix.
     * 
     * <p>
     * Note: if the paint contains a maskfilter that generates a mask which extends beyond the bitmap's original
     * width/height (e.g. BlurMaskFilter), then the bitmap will be drawn as if it were in a Shader with CLAMP mode. Thus
     * the color outside of the original width/height will be the edge color replicated.
     * 
     * <p>
     * If the bitmap and canvas have different densities, this function will take care of automatically scaling the
     * bitmap to draw at the same density as the canvas.
     * 
     * Unsupported; ignored if called
     * 
     * @param bitmap
     *            The bitmap to be drawn
     * @param left
     *            The position of the left side of the bitmap being drawn
     * @param top
     *            The position of the top side of the bitmap being drawn
     * @param paint
     *            The paint used to draw the bitmap (may be null)
     */
    public void drawBitmap(Bitmap bitmap, float left, float top, Paint paint)
    {
        /*
        DXFImage image = new DXFImage(bitmap);
         
        
        // Add the image to the document and the page resource-reference table
        page.getDXFDocument().add(image);
        page.addImageResource(image);
        page.addToProcset("/ImageC");
        
        startImage(left, top, bitmap.getWidth(), bitmap.getHeight());
        pw.println(image.getName() + " Do");
        endImage();
        */
    }
    
    
    
    /**
     * Draw the specified bitmap, scaling/translating automatically to fill the destination rectangle. If the source
     * rectangle is not null, it specifies the subset of the bitmap to draw.
     * 
     * <p>
     * Note: if the paint contains a maskfilter that generates a mask which extends beyond the bitmap's original
     * width/height (e.g. BlurMaskFilter), then the bitmap will be drawn as if it were in a Shader with CLAMP mode. Thus
     * the color outside of the original width/height will be the edge color replicated.
     * 
     * <p>
     * This function <em>ignores the density associated with the bitmap</em>. This is because the source and destination
     * rectangle coordinate spaces are in their respective densities, so must already have the appropriate scaling
     * factor applied.
     * 
     * Unsupported; ignored if called
     * 
     * @param bitmap
     *            The bitmap to be drawn
     * @param src
     *            May be null. The subset of the bitmap to be drawn
     * @param dst
     *            The rectangle that the bitmap will be scaled/translated to fit into
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(Bitmap bitmap, Rect src, RectF dst, Paint paint)
    {
        /*
        DXFImage image;
        
        if (dst == null)
        {
            throw new NullPointerException();
        }
        
        if (src != null)
        {
            image = new DXFImage(bitmap, src.left, src.top, src.width(), src.height());
        }
        else
        {
            image = new DXFImage(bitmap);
        }
        
        // Add the image to the document and the page resource-reference table
        page.getDXFDocument().add(image);
        page.addImageResource(image);
        page.addToProcset("/ImageC");
        
        startImage(dst.left, dst.top, dst.width(), dst.height());
        pw.println(image.getName() + " Do");
        endImage();
        */
    }
    
    
    
    /**
     * Draw the specified bitmap, scaling/translating automatically to fill the destination rectangle. If the source
     * rectangle is not null, it specifies the subset of the bitmap to draw.
     * 
     * <p>
     * Note: if the paint contains a maskfilter that generates a mask which extends beyond the bitmap's original
     * width/height (e.g. BlurMaskFilter), then the bitmap will be drawn as if it were in a Shader with CLAMP mode. Thus
     * the color outside of the original width/height will be the edge color replicated.
     * 
     * <p>
     * This function <em>ignores the density associated with the bitmap</em>. This is because the source and destination
     * rectangle coordinate spaces are in their respective densities, so must already have the appropriate scaling
     * factor applied.
     * 
     * Unsupported; ignored if called
     * 
     * @param bitmap
     *            The bitmap to be drawn
     * @param src
     *            May be null. The subset of the bitmap to be drawn
     * @param dst
     *            The rectangle that the bitmap will be scaled/translated to fit into
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(Bitmap bitmap, Rect src, Rect dst, Paint paint)
    {  
        if (dst == null)
        {
            throw new NullPointerException();
        }
        
        drawBitmap(bitmap, src, new RectF(dst), paint);
    }
    
    
    
    /**
     * Treat the specified array of colors as a bitmap, and draw it. This gives the same result as first creating a
     * bitmap from the array, and then drawing it, but this method avoids explicitly creating a bitmap object which can
     * be more efficient if the colors are changing often.
     * 
     * Unsupported; ignored if called
     * 
     * @param colors
     *            Array of colors representing the pixels of the bitmap
     * @param offset
     *            Offset into the array of colors for the first pixel
     * @param stride
     *            The number of colors in the array between rows (must be >= width or <= -width).
     * @param x
     *            The X coordinate for where to draw the bitmap
     * @param y
     *            The Y coordinate for where to draw the bitmap
     * @param width
     *            The width of the bitmap
     * @param height
     *            The height of the bitmap
     * @param hasAlpha
     *            True if the alpha channel of the colors contains valid values. If false, the alpha byte is ignored
     *            (assumed to be 0xFF for every pixel).
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(int[] colors, int offset, int stride, float x, float y, int width, int height, boolean hasAlpha, Paint paint)
    {
        
        // check for valid input
        if (width < 0)
        {
            throw new IllegalArgumentException("width must be >= 0");
        }
        if (height < 0)
        {
            throw new IllegalArgumentException("height must be >= 0");
        }
        if (Math.abs(stride) < width)
        {
            throw new IllegalArgumentException("abs(stride) must be >= width");
        }
        int lastScanline = offset + (height - 1) * stride;
        int length = colors.length;
        if (offset < 0 || (offset + width > length) || lastScanline < 0 || (lastScanline + width > length))
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        // quick escape if there's nothing to draw
        if (width == 0 || height == 0)
        {
            return;
        }
        
        // Be lazy - just create the damn Bitmap...
        Bitmap bitmap = Bitmap.createBitmap(colors, offset, stride, width, height, Bitmap.Config.ARGB_8888);
        drawBitmap(bitmap, x, y, paint);
        
    }
    
    
    
    /**
     * Legacy version of drawBitmap(int[] colors, ...) that took ints for x,y
     * 
     * Unsupported; ignored if called
     */
    public void drawBitmap(int[] colors, int offset, int stride, int x, int y, int width, int height, boolean hasAlpha, Paint paint)
    {
        // call through to the common float version
        drawBitmap(colors, offset, stride, (float) x, (float) y, width, height, hasAlpha, paint);
    }
    
    
    
    /**
     * Draw the bitmap using the specified matrix.
     * 
     * Unsupported; ignored if called
     * 
     * @param bitmap
     *            The bitmap to draw
     * @param matrix
     *            The matrix used to transform the bitmap when it is drawn
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmap(Bitmap bitmap, Matrix transformMatrix, Paint paint)
    {
        /*
        DXFImage image = new DXFImage(bitmap);
        
        // Add the image to the document and the page resource-reference table
        page.getDXFDocument().add(image);
        page.addImageResource(image);
        page.addToProcset("/ImageC");
        
        // create matrix copy since we're going to change it
        Matrix matrix = new Matrix(transformMatrix);
        
        // need to shift down by the bitmap height since Java coords are for top left but
        // DXF uses bottom left
        matrix.preTranslate(0, bitmap.getHeight());
        
        // need to scale since the DXF image model assumes image fits
        // in a unit square; so multiply by the width and height
        matrix.preScale(bitmap.getWidth(), bitmap.getHeight());
        
        // startImage(matrix, bitmap.getWidth(), bitmap.getHeight());
        startImage(matrix);
        pw.println(image.getName() + " Do");
        endImage();
        */
    }
    
    
    
    /**
     * Draw the bitmap through the mesh, where mesh vertices are evenly distributed across the bitmap. There are
     * meshWidth+1 vertices across, and meshHeight+1 vertices down. The verts array is accessed in row-major order, so
     * that the first meshWidth+1 vertices are distributed across the top of the bitmap from left to right.
     * 
     * This method stretches the supplied bitmap across the specified mesh, which can be used to create 3D effects in
     * which the bitmap appears to be applied to the surface for which the mesh is the projection on the plane of the
     * page.
     * 
     * Unsupported; ignored if called
     * 
     * @param bitmap
     *            The bitmap to draw using the mesh
     * @param meshWidth
     *            The number of columns in the mesh. Nothing is drawn if this is 0
     * @param meshHeight
     *            The number of rows in the mesh. Nothing is drawn if this is 0
     * @param verts
     *            Array of x,y pairs, specifying where the mesh should be drawn. There must be at least (meshWidth+1) *
     *            (meshHeight+1) * 2 + meshOffset values in the array
     * @param vertOffset
     *            Number of verts elements to skip before drawing
     * @param colors
     *            May be null. Specifies a color at each vertex, which is interpolated across the cell, and whose values
     *            are multiplied by the corresponding bitmap colors. If not null, there must be at least (meshWidth+1) *
     *            (meshHeight+1) + colorOffset values in the array.
     * @param colorOffset
     *            Number of color elements to skip before drawing
     * @param paint
     *            May be null. The paint used to draw the bitmap
     */
    public void drawBitmapMesh(Bitmap bitmap, int meshWidth, int meshHeight, float[] verts, int vertOffset, int[] colors, int colorOffset, Paint paint)
    {
        /*
        if ((meshWidth | meshHeight | vertOffset | colorOffset) < 0)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        if (meshWidth == 0 || meshHeight == 0)
        {
            return;
        }
        int count = (meshWidth + 1) * (meshHeight + 1);
        // we mul by 2 since we need two floats per vertex
        checkRange(verts.length, vertOffset, count * 2);
        if (colors != null)
        {
            // no mul by 2, since we need only 1 color per vertex
            checkRange(colors.length, colorOffset, count);
        }
        
        // OK get clever - draw a series of bitmaps for the triangles of the mesh,
        // by drawing them as clipped paralellograms
        for (int i = 0; i < meshWidth; i++)
        {
            for (int j = 0; j < meshHeight; j++)
            {
                int topLeftIndex = vertOffset + 2 * (i + j * (meshWidth + 1));
                int topRightIndex = vertOffset + 2 * ((i + 1) + j * (meshWidth + 1));
                int bottomLeftIndex = vertOffset + 2 * (i + (j + 1) * (meshWidth + 1));
                int bottomRightIndex = vertOffset + 2 * ((i + 1) + (j + 1) * (meshWidth + 1));
                
                // the points in our quadrilateral
                PointF topLeft = new PointF(verts[topLeftIndex], verts[topLeftIndex + 1]);
                PointF topRight = new PointF(verts[topRightIndex], verts[topRightIndex + 1]);
                PointF bottomLeft = new PointF(verts[bottomLeftIndex], verts[bottomLeftIndex + 1]);
                PointF bottomRight = new PointF(verts[bottomRightIndex], verts[bottomRightIndex + 1]);
                
                // subbitmap image
                int subbitmapWidth = bitmap.getWidth() / meshWidth;
                int subbitmapHeight = bitmap.getHeight() / meshHeight;
                DXFImage image = new DXFImage(bitmap, i * subbitmapWidth, j * subbitmapHeight, subbitmapWidth, subbitmapHeight);
                
                // Add the image to the document and the page resource-reference table
                page.getDXFDocument().add(image);
                page.addImageResource(image);
                page.addToProcset("/ImageC");
                
                // now create matrix to draw this bitmap into parallelogram containing top triangle
                Matrix matrix = new Matrix();
                
                // float[] topSrcBasis = {1, 0, 0, 1};
                // float[] topDstBasis = {bottomRight.x - bottomLeft.x, bottomRight.y - bottomLeft.y, topLeft.x -
                // bottomLeft.x, topLeft.y - bottomLeft.y};
                // debugArray("topSrcBasis", topSrcBasis);
                // debugArray("topDstBasis", topDstBasis);
                float[] topDstPoints = { topLeft.x, topLeft.y, topRight.x, topRight.y, bottomRight.x, bottomRight.y };
                
                // set matrix coefficients based on basis vectors and base point for translation; note that base point
                // is the completion of the top-right parallelogram - it's not one of the 4 points
                float[] matrixCoefs = new float[9];
                matrixCoefs[Matrix.MSCALE_X] = (topRight.x - topLeft.x);
                matrixCoefs[Matrix.MSKEW_X] = -(topRight.y - topLeft.y);
                matrixCoefs[Matrix.MSCALE_Y] = -(topRight.y - bottomRight.y);
                matrixCoefs[Matrix.MSKEW_Y] = (topRight.x - bottomRight.x);
                matrixCoefs[Matrix.MTRANS_X] = topLeft.x + (bottomRight.x - topRight.x);
                matrixCoefs[Matrix.MTRANS_Y] = topLeft.y + (bottomRight.y - topRight.y);
                matrixCoefs[Matrix.MPERSP_0] = 0;
                matrixCoefs[Matrix.MPERSP_1] = 0;
                matrixCoefs[Matrix.MPERSP_2] = 1;
                
                matrix.setValues(matrixCoefs);
                
                // draw the image, but clipping before to get just top right triangle;
                // save the state so we can undo the clipping
                save();
                clipPoly(topDstPoints);
                startImage(matrix);
                pw.println(image.getName() + " Do");
                endImage();
                restore();
                
                // now set matrix to draw this bitmap into parallelogram containing bottom-right triangle
                
                // we use the mapping of the fundamental unit basis to the sides of the triangle to create the matrix
                // scaling/skewing coefficients, and the location of the bottom-left point for the translation
                // float[] bottomSrcBasis = {1, 0, 0, 1};
                // float[] bottomDstBasis = {bottomRight.x - bottomLeft.x, bottomRight.y - bottomLeft.y, topLeft.x -
                // bottomLeft.x, topLeft.y - bottomLeft.y};
                // debugArray("bottomSrcBasis", bottomSrcBasis);
                // debugArray("bottomDstBasis", bottomDstBasis);
                float[] bottomDstPoints = { bottomLeft.x, bottomLeft.y, topLeft.x, topLeft.y, bottomRight.x, bottomRight.y };
                
                matrixCoefs = new float[9];
                matrixCoefs[Matrix.MSCALE_X] = (bottomRight.x - bottomLeft.x);
                matrixCoefs[Matrix.MSKEW_X] = -(bottomRight.y - bottomLeft.y);
                matrixCoefs[Matrix.MSCALE_Y] = -(topLeft.y - bottomLeft.y);
                matrixCoefs[Matrix.MSKEW_Y] = (topLeft.x - bottomLeft.x);
                matrixCoefs[Matrix.MTRANS_X] = bottomLeft.x;
                matrixCoefs[Matrix.MTRANS_Y] = bottomLeft.y;
                matrixCoefs[Matrix.MPERSP_0] = 0;
                matrixCoefs[Matrix.MPERSP_1] = 0;
                matrixCoefs[Matrix.MPERSP_2] = 1;
                
                matrix.setValues(matrixCoefs);
                
                // draw the image, but clipping before to get just bottom left triangle;
                // save the state so we can restore the clip
                save();
                clipPoly(bottomDstPoints);
                startImage(matrix);
                pw.println(image.getName() + " Do");
                endImage();
                restore();
                
            }
        }
        */
        
    }
    
    
    
    /**
     * The VertexMode determines how the subsequent points in the sequence of vertices are interpreted.
     * 
     * - TRIANGLES In this mode each triple of points (6 successive coordinates) is treated as a separate triangle. This
     * is the most general mode; no assumption is made about triangle adjacency.
     * 
     * - TRIANGLE_STRIP In this mode each successive vertex is combined with the two before it to form a new triangle,
     * giving a "strip" of triangles where each shares an edge with the triangle before it
     * 
     * - TRIANGLE_FAN In this mode the first point is assumed to be the center vertex for all of the triangles, with
     * subsequent points defining the "outer" vertices of the fan arranged radially around the center. The triangles are
     * thus "pie wedges" with the first point at the center of the pie, and each "wedge" sharing an edge and vertex.
     * Note that the first and last outer vertices are not automatically joined.
     * 
     * @author jsevy
     *
     */
    /*
     * public enum VertexMode { TRIANGLES(0), TRIANGLE_STRIP(1), TRIANGLE_FAN(2);
     * 
     * VertexMode(int nativeInt) { this.nativeInt = nativeInt; }
     * 
     * public final int nativeInt; }
     */
    
    /**
     * Draw the array of vertices, interpreted as triangles (based on mode). The verts array is required, and specifies
     * the x,y pairs for each vertex. The color array specifies the color at each vertex, and the colors at each vertex
     * are interpolated across the corresponding triangle in a gradient. The indices and texs arrays are currently
     * ignored.
     * <p>
     * The VertexMode determines how the subsequent points in the sequence of vertices are interpreted.
     * <ul>
     * <li>
     * TRIANGLES In this mode each triple of points (6 successive coordinates) is treated as a separate triangle. This
     * is the most general mode; no assumption is made about triangle adjacency.</li>
     * <li>
     * TRIANGLE_STRIP In this mode each successive vertex is combined with the two before it to form a new triangle,
     * giving a "strip" of triangles where each shares an edge with the triangle before it</li>
     * <li>
     * TRIANGLE_FAN In this mode the first point is assumed to be the center vertex for all of the triangles, with
     * subsequent points defining the "outer" vertices of the fan arranged radially around the center. The triangles are
     * thus "pie wedges" with the first point at the center of the pie, and each "wedge" sharing an edge and vertex.
     * Note that the first and last outer vertices are not automatically joined.</li>
     * </ul>
     * </p>
     * 
     * Unsupported; ignored if called
     * 
     * @param mode
     *            How to interpret the array of vertices
     * @param vertexCount
     *            The number of values in the vertices array (and corresponding texs and colors arrays if non-null).
     *            Each logical vertex is two values (x, y), vertexCount must be a multiple of 2.
     * @param verts
     *            Array of vertices for the mesh
     * @param vertOffset
     *            Number of values in the verts to skip before drawing.
     * @param texs
     *            Currently ignored (may be null)
     * @param texOffset
     *            Currently ignored
     * @param colors
     *            Specifies a color for each vertex, to be interpolated across the triangle.
     * @param colorOffset
     *            Number of values in colors to skip before drawing.
     * @param indices
     *            Currently ignored (may be null)
     * @param indexCount
     *            Currently ignored
     * @param paint
     *            Currently ignored (may be null)
     */
    public void drawVertices(Canvas.VertexMode mode, int vertexCount, float[] verts, int vertOffset, float[] texs, int texOffset, int[] colors, int colorOffset, short[] indices, int indexOffset, int indexCount, Paint paint)
    {
        /*
        checkRange(verts.length, vertOffset, vertexCount);
        if (texs != null)
        {
            checkRange(texs.length, texOffset, vertexCount);
        }
        if (colors != null)
        {
            checkRange(colors.length, colorOffset, vertexCount / 2);
        }
        if (indices != null)
        {
            checkRange(indices.length, indexOffset, indexCount);
        }
        
        // convert vertices to pdf coordinates
        float[] transformedVerts = new float[verts.length];
        graphicsMatrix.mapPoints(transformedVerts, verts);
        
        // create a new Gouraud pattern object adn add it to the page and document
        DXFShadingGouraudFreeform shading = new DXFShadingGouraudFreeform(mode, vertexCount, transformedVerts, vertOffset, colors, colorOffset);
        
        // Add the image to the document and the page resource-reference table
        page.getDXFDocument().add(shading);
        page.addShadingResource(shading);
        
        // stroke the shading
        pw.println(shading.getName() + " sh");
        
        // for debugging shading contents
        pw.println("%" + shading.getStreamDataString());
        */
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), using the typeface and size in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param x
     *            The x-coordinate of the origin of the text being drawn
     * @param y
     *            The y-coordinate of the origin of the text being drawn
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(char[] text, int index, int count, float x, float y, Paint paint)
    {
        if ((index | count | (index + count) | (text.length - index - count)) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        
        String textString = new String(text, index, count);
        drawText(textString, x, y, paint);
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), using the typeface and size in the specified paint.
     * 
     * @param string
     *            The text to be drawn
     * @param x
     *            The x-coordinate of the origin of the text being drawn
     * @param y
     *            The y-coordinate of the origin of the text being drawn
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(String string, float x, float y, Paint paint)
    {
        DXFStyle style = addFontStyle(paint);
        
        // transform the coordinates using the current graphics matrix
        float[] coords = new float[]{x, y};
        graphicsMatrix.mapPoints(coords);
        
        // get the rotation and oblique angles from the current graphics matrix
        // by mapping the unit vectors
        float[] unitVectorICoords = new float[]{1, 0};
        float[] unitVectorJCoords = new float[]{0, 1};
        float[] unitVectorITransformedCoords = new float[2];
        float[] unitVectorJTransformedCoords = new float[2];
        graphicsMatrix.mapVectors(unitVectorITransformedCoords, unitVectorICoords);
        graphicsMatrix.mapVectors(unitVectorJTransformedCoords, unitVectorJCoords);
        
        RealPoint unitVectorITransformed = new RealPoint(unitVectorITransformedCoords[0], unitVectorITransformedCoords[1], 0);
        RealPoint unitVectorI = new RealPoint(unitVectorICoords[0], unitVectorICoords[1], 0);
        RealPoint unitVectorJTransformed = new RealPoint(unitVectorJTransformedCoords[0], unitVectorJTransformedCoords[1], 0);
        
        // compute rotation and oblique angles, using cross product to determine sign
        float rotationAngle = RealPoint.angleBetween(unitVectorI, unitVectorITransformed);
        if (unitVectorITransformed.y < 0)
            rotationAngle = -rotationAngle;
        
        float obliqueAngle = RealPoint.angleBetween(unitVectorITransformed, unitVectorJTransformed) - (float)Math.PI/2;
        if (unitVectorJTransformed.x * unitVectorITransformed.y - unitVectorJTransformed.y * unitVectorITransformed.x < 0)
            obliqueAngle = -obliqueAngle;
        
        // convert to degrees
        rotationAngle = (180/(float)Math.PI) * rotationAngle;
        obliqueAngle = (180/(float)Math.PI) * obliqueAngle;
        
        
        DXFText text = new DXFText(string, new RealPoint(coords[0], coords[1], 0), rotationAngle, obliqueAngle, style, paint);
        dxfDocument.addEntity(text);
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), using the typeface and size in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param start
     *            The index of the first character in text to draw
     * @param end
     *            (end - 1) is the index of the last character in text to draw
     * @param x
     *            The x-coordinate of the origin of the text being drawn
     * @param y
     *            The y-coordinate of the origin of the text being drawn
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(String text, int start, int end, float x, float y, Paint paint)
    {
        if ((start | end | (end - start) | (text.length() - end)) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        
        String textString = text.substring(start, end);
        drawText(textString, x, y, paint);
    }
    
    
    
    /**
     * Draw the specified range of text, specified by start/end, with its origin at (x,y), using the typeface and size
     * in the specified paint.
     * 
     * @param text
     *            The text to be drawn
     * @param start
     *            The index of the first character in text to draw
     * @param end
     *            (end - 1) is the index of the last character in text to draw
     * @param x
     *            The x-coordinate of origin for where to draw the text
     * @param y
     *            The y-coordinate of origin for where to draw the text
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawText(CharSequence text, int start, int end, float x, float y, Paint paint)
    {
        char[] buf = new char[end - start];
        TextUtils.getChars(text, start, end, buf, 0);
        drawText(buf, 0, end - start, x, y, paint);
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     * 
     * @param text
     *            the text to render
     * @param index
     *            the start of the text to render
     * @param count
     *            the count of chars to render
     * @param contextIndex
     *            the start of the context for shaping. Must be no greater than index.
     * @param contextCount
     *            the number of characters in the context for shaping. ContexIndex + contextCount must be no less than
     *            index + count.
     * @param x
     *            the x position at which to draw the text
     * @param y
     *            the y position at which to draw the text
     * @param dir
     *            the run direction, either {@link #DIRECTION_LTR} or {@link #DIRECTION_RTL}.
     * @param paint
     *            the paint
     * @hide
     */
    public void drawTextRun(char[] text, int index, int count, int contextIndex, int contextCount, float x, float y, int dir, Paint paint)
    {
        
        if (text == null)
        {
            throw new NullPointerException("text is null");
        }
        if (paint == null)
        {
            throw new NullPointerException("paint is null");
        }
        if ((index | count | text.length - index - count) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        if (dir != DIRECTION_LTR && dir != DIRECTION_RTL)
        {
            throw new IllegalArgumentException("unknown dir: " + dir);
        }
        
        throw new UnsupportedOperationException("Operation drawTextRun not currently supported in DXFCanvas");
        
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     * 
     * @param text
     *            the text to render
     * @param start
     *            the start of the text to render. Data before this position can be used for shaping context.
     * @param end
     *            the end of the text to render. Data at or after this position can be used for shaping context.
     * @param x
     *            the x position at which to draw the text
     * @param y
     *            the y position at which to draw the text
     * @param dir
     *            the run direction, either 0 for LTR or 1 for RTL.
     * @param paint
     *            the paint
     * @hide
     */
    public void drawTextRun(CharSequence text, int start, int end, int contextStart, int contextEnd, float x, float y, int dir, Paint paint)
    {
        
        if (text == null)
        {
            throw new NullPointerException("text is null");
        }
        if (paint == null)
        {
            throw new NullPointerException("paint is null");
        }
        if ((start | end | end - start | text.length() - end) < 0)
        {
            throw new IndexOutOfBoundsException();
        }
        
        // int flags = dir == 0 ? 0 : 1;
        
        // To do
        throw new UnsupportedOperationException("Operation drawTextRun not currently supported in DXFCanvas");
        
    }
    
    
    
    /**
     * Draw the text in the array, with each character's origin specified by the pos array, using the typeface and size
     * in the specified Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param index
     *            The index of the first character to draw
     * @param count
     *            The number of characters to draw, starting from index.
     * @param pos
     *            Array of [x,y] positions, used to position each character
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawPosText(char[] text, int index, int count, float[] pos, Paint paint)
    {
        if (index < 0 || index + count > text.length || count * 2 > pos.length)
        {
            throw new IndexOutOfBoundsException();
        }
        
        // add text style to STYLE table if not already there
        DXFStyle style = addFontStyle(paint);
        
        for (int i = index; i < index + count; i++)
        {
            char[] data = {text[i]};
            String charString = new String(data);
            DXFText dxfText = new DXFText(charString, new RealPoint(pos[2 * i], -pos[2 * i + 1], 0), style, paint);
            dxfDocument.addEntity(dxfText);
        }
        
    }
    
    
    
    /**
     * Draw the text in the array, with each character's origin specified by the pos array, using the typeface and size
     * in the specified Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param pos
     *            Array of [x,y] positions, used to position each character
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawPosText(String text, float[] pos, Paint paint)
    {
        if (text.length() * 2 > pos.length)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        
        char[] textChars = text.toCharArray();
        drawPosText(textChars, 0, textChars.length, pos, paint);
    }
    
    
    
    
    /**
     * Draw the text, with origin at (x,y), along the specified path, using the typeface and size in the specified
     * Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param path
     *            The path the text should follow for its baseline
     * @param hOffset
     *            The distance along the path to add to the text's starting position
     * @param vOffset
     *            The distance above(-) or below(+) the path to position the text
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawTextOnPath(char[] text, int index, int count, Path path, float hOffset, float vOffset, Paint paint)
    {
        if (index < 0 || index + count > text.length)
        {
            throw new ArrayIndexOutOfBoundsException();
        }
        
        drawTextOnPath(new String(text, index, count), path, hOffset, vOffset, paint);
    }
    
    
    
    /**
     * Draw the text, with origin at (x,y), along the specified path, using the typeface and size in the specified
     * Paint.
     * 
     * @param text
     *            The text to be drawn
     * @param path
     *            The path the text should follow for its baseline
     * @param hOffset
     *            The distance along the path to add to the text's starting position
     * @param vOffset
     *            The distance above(-) or below(+) the path to position the text
     * @param paint
     *            The paint used for the text (e.g. color, size, style)
     */
    public void drawTextOnPath(String text, Path path, float hOffset, float vOffset, Paint paint)
    {
        
        if (text.length() > 0)
        {
            // add text style to STYLE table if not already there
            DXFStyle style = addFontStyle(paint);
            
            // Use PathMeasure to get positions along the path
            PathMeasure pathMeasure = new PathMeasure(path, false);
            float pathLength = pathMeasure.getLength();
            if ((hOffset > pathLength) || (pathLength == 0))
                return;
            
            float[] textWidths = new float[text.length()];
            paint.getTextWidths(text, textWidths);
            
            float[] pos = new float[2];
            float[] tan = new float[2];
            float length = hOffset;
            
            // get starting position of text along path, and offset by
            // vOffset perpendicular to path
            pathMeasure.getPosTan(length, pos, tan);
            pos[0] += vOffset * -tan[1];
            pos[1] += vOffset * tan[0];
            
            // get text rotation
            float rot = (float) (Math.acos(tan[0]) * (180 / Math.PI)) * Math.signum(tan[1]);
            
            for (int i = 0; i < text.length(); i++)
            {
                // draw text at pos along curve
                DXFText dxfText = new DXFText(text.substring(i, i+1), new RealPoint(pos[0], -pos[1], 0), -rot, style, paint);
                dxfDocument.addEntity(dxfText);
                
                // get next position info
                length += textWidths[i];
                pathMeasure.getPosTan(length, pos, tan);
                pos[0] += vOffset * -tan[1];
                pos[1] += vOffset * tan[0];
                rot = (float) (Math.acos(tan[0]) * (180 / Math.PI)) * Math.signum(tan[1]);
            }
        }
        
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     * 
     * @param picture
     *            The picture to be drawn
     */
    public void drawPicture(Picture picture)
    {
        /*
         * picture.endRecording();
         * 
         * // save state save(); picture.draw(this); restore();
         */
        
        throw new UnsupportedOperationException("Operation drawPicture not supported in DXFCanvas");
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     */
    public void drawPicture(Picture picture, RectF dst)
    {
        /*
         * save(); translate(dst.left, dst.top); if (picture.getWidth() > 0 && picture.getHeight() > 0) {
         * scale(dst.width() / picture.getWidth(), dst.height() / picture.getHeight()); } drawPicture(picture);
         * restore();
         */
        
        throw new UnsupportedOperationException("Operation drawPicture not supported in DXFCanvas");
    }
    
    
    
    /**
     * Currently unsupported; throws UnsupportedOperationException if called.
     */
    public void drawPicture(Picture picture, Rect dst)
    {
        /*
         * save(); translate(dst.left, dst.top); if (picture.getWidth() > 0 && picture.getHeight() > 0) { scale((float)
         * dst.width() / picture.getWidth(), (float) dst.height() / picture.getHeight()); } drawPicture(picture);
         * restore();
         */
        
        throw new UnsupportedOperationException("Operation drawPicture not supported in DXFCanvas");
    }
    
    
    
    private boolean takesCirclesToCircles(Matrix transformation)
    {
    	// map basis vectors
        float[] basisVects = {1, 0, 0, 1};
        graphicsMatrix.mapVectors(basisVects);
        float c11 = basisVects[0];
        float c12 = basisVects[2];
        float c21 = basisVects[1];
        float c22 = basisVects[3];
        
        RealPoint i = new RealPoint(1, 0, 0);
        RealPoint j = new RealPoint(0, 1, 0);
        
        RealPoint u = new RealPoint(c11, c21, 0);
        RealPoint v = new RealPoint(c12, c22, 0);
        
        if ((RealPoint.magnitude(u) == RealPoint.magnitude(v)) && (RealPoint.dotProduct(u, v) == 0))
        {
        	return true;
        }
        else
        {
        	return false;
        }
    }
    
    
    private boolean preservesAxisOrientations(Matrix transformation)
    {
    	// see if the supplied transformation matrix is a pure scaling plus translation
    	
    	// map basis vectors
        float[] basisVects = {1, 0, 0, 1};
        transformation.mapVectors(basisVects);
        float c11 = basisVects[0];
        float c12 = basisVects[2];
        float c21 = basisVects[1];
        float c22 = basisVects[3];
        
        RealPoint i = new RealPoint(1, 0, 0);
        RealPoint j = new RealPoint(0, 1, 0);
        
        RealPoint u = RealPoint.mapVector(transformation, i);
        RealPoint v = RealPoint.mapVector(transformation, j);
        
        float uMag = RealPoint.magnitude(u);
        float vMag = RealPoint.magnitude(v);
        float angleIU = RealPoint.angleBetween(i, u);
        float angleJV = RealPoint.angleBetween(j, v);
        
        if ((uMag != 0) && (vMag != 0) && (angleIU == 0) && (angleJV == 0))
        {
            return true;
        }
        else
        {
        	return false;
        }
    }
    
    
    
}
