/*
 * ADXF Library
 * 
 *   Copyright (C) 2016, Jonathan Sevy <jsevy@cs.drexel.edu>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.adxf;



/**
 * Class representing a DXF document, which owns a DXFCanvas on which drawing commands can be made. The document's
 * toDXFString method can then be called to generate the DXF text corresponding to the content of the canvas.
 * 
 * @author jsevy
 *
 */
public class DXFDocument
{
    private DXFSection header = new DXFSection("HEADER");
    private DXFSection classes = new DXFSection("CLASSES");
    private DXFSection tables = new DXFSection("TABLES");
    private DXFSection blocks = new DXFSection("BLOCKS");
    private DXFSection entities = new DXFSection("ENTITIES");
    private DXFSection objects = new DXFSection("OBJECTS");
    
    private String documentComment = "";
    
    private DXFCanvas canvas;
    
    
    /**
     * Create a new DXFDocument
     */
    public DXFDocument()
    {
        this("");
    }
    
    
    /**
     * Create a new DXF document with the specified comment in its header
     * 
     * @param documentComment
     */
    public DXFDocument(String documentComment)
    {
        this.documentComment = documentComment;
        this.canvas = new DXFCanvas(this);
    }
    
    
    /**
     * Get the DXFCanvas associated with this document for use with standard Canvas drawing operations to generate
     * a DXF text representation.
     * 
     * @return	The DXFCanvas associated with this document, on which standard Android canvas drawing calls can be made
     */
    public DXFCanvas getCanvas()
    {
        return this.canvas;
    }
    
    
    /**
     * Return the DXF text associated with this DXF document. This includes the header,
     * classes, tables, blocks, entities and objects sections, populated with content
     * generated by graphics calls on the associated DXFCanvas.
     * 
     * @return The DXF text associated with this document.
     */
    public String toDXFString()
    {
        String returnString = new String();
        
        returnString += "999\n" + documentComment + "\n";
        
        returnString += header.toDXFString();
        returnString += classes.toDXFString();
        returnString += tables.toDXFString();
        returnString += blocks.toDXFString();
        returnString += entities.toDXFString();
        returnString += objects.toDXFString();
        
        // end-of-file marker
        returnString += "0\nEOF\n";
        
        return returnString;
    }
    
    
    /**
     * Utility method used by the associated DXFCanvas object.
     * 
     * @param table A DXFTable instance
     */
    public void addTable(DXFTable table)
    {
        tables.add(table);
    }
    
    
    /**
     * Utility method used by the associated DXFCanvas object.
     * 
     * @param entity A DXFEntity instance
     */
    public void addEntity(DXFEntity entity)
    {
        entities.add(entity);
    }
    
    
    /**
     * Utility method used by the associated DXFCanvas object.
     * 
     * @param style A DXFStyle instance
     */
    public DXFStyle addStyle(DXFStyle style)
    {
        // first see if style already represented in tables; if so, return it
        for(int i = 0; i < tables.size(); i++)
        {
            DXFTable table = (DXFTable)tables.elementAt(i);
            if (table.name.equals("STYLE"))
            {
                // see if we have the style in the table
                int index = table.indexOf(style);
                if (index > 0)
                {
                    return (DXFStyle)table.elementAt(index);
                }
            }
        }
        
        
        // didn't find it; add to one of the STYLE tables, adding one if necessary, and return the style passed in
        DXFTable styleTable = null;
        for(int i = 0; i < tables.size(); i++)
        {
            DXFTable table = (DXFTable)tables.elementAt(i);
            if (table.name.equals("STYLE"))
            {
                styleTable = table;
                break;
            }
        }
        
        if (styleTable == null)
        {
            styleTable = new DXFTable("STYLE");
            tables.add(styleTable);
        }
        
        styleTable.add(style);
        
        return style;
    }
    
}