/*
 * ADXF Library
 * 
 *   Copyright (C) 2016, Jonathan Sevy <jsevy@cs.drexel.edu>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.adxf;

import android.graphics.Paint;


/**
 * Class representing an elliptical arc.
 * 
 * @author jsevy
 *
 */
public class DXFEllipse implements DXFEntity
{
    private RealPoint center;
    private RealPoint majorAxisEndpoint;    // relative to center
    private double axisRatio;                // minor to major
    private double startParameter;	// parameter t corresponding to the start point of the arc represented as (a*cos(t), b*sin(t))
    private double endParameter;		// parameter t corresponding to the end point of the arc represented as (a*cos(t), b*sin(t))
    private Paint paint;
    
    
    /**
     * Create an elliptical arc corresponding to the specified parameters. Note that this is the most general case
     * of an ellipse, since the major and minor axes need not be parallel to the x or y axes.
     * 
     * @param center			The center of the ellipse
     * @param majorAxisEndpoint	The endpoint of the major axis, relative to the center of the ellipse
     * @param axisRatio			The ratio between the length of the minor axis and the length of the major axis
     * @param startParameter	Parameter t corresponding to the start point of the arc represented as (a*cos(t), b*sin(t)),
     * 							rotated so that t = 0 corresponds to the major axis endpoint.
     * @param endParameter	    Parameter t corresponding to the end point of the arc represented as (a*cos(t), b*sin(t)),
     *                          rotated so that t = 0 corresponds to the major axis endpoint.
     * @param paint				Android Paint object containing drawing information (color, line width, etc.)
     */
    public DXFEllipse(RealPoint center, RealPoint majorAxisEndpoint, double axisRatio, double startParameter, double endParameter, Paint paint)
    {
        this.center = new RealPoint(center);
        this.majorAxisEndpoint = new RealPoint(majorAxisEndpoint);
        this.axisRatio = axisRatio;
        this.startParameter = startParameter;
        this.endParameter = endParameter;
        this.paint = new Paint(paint);
    }
    
    
    
    
    /**
     * Implementation of DXFObject interface method; creates DXF text representing the elliptical arc.
     */
    public String toDXFString()
    {
        String result = "0\nELLIPSE\n";
        
        // center
        result += "10\n" + center.x + "\n";
        result += "20\n" + center.y + "\n";
        result += "30\n" + center.z + "\n";
       
        // major axis endpoint relative to center
        result += "11\n" + majorAxisEndpoint.x + "\n";
        result += "21\n" + majorAxisEndpoint.y + "\n";
        result += "31\n" + majorAxisEndpoint.z + "\n";
       
        // radius ratio
        result += "40\n" + axisRatio + "\n";
       
        // start/end angles
        result += "41\n" + startParameter + "\n";
        result += "42\n" + endParameter + "\n";
       
        // add thickness
        //result += "370\n" + paint.getStrokeWidth() + "\n";

        // add color number
        result += "62\n" + DXFColor.getClosestDXFColor(paint.getColor()) + "\n";
        
        return result;
    }
}