/*
 * ADXF Library
 * 
 *   Copyright (C) 2016, Jonathan Sevy <jsevy@cs.drexel.edu>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.adxf;

import android.graphics.Typeface;



/**
 * Class representing a text style for use in a DXF document. Encapsulates text parameters (font, style), 
 * and associates a "name" (just an integer) for use in text entities which use that style.
 * @author jsevy
 *
 */
public class DXFStyle implements DXFObject
{
    private int name;
    private String dxfFontName;
    private Typeface androidFont;
    
    private static int highName = 0;
    
    
    
    /**
     * Create a style object corresponding to the supplied Android typeface object.
     * 
     * @param androidFont	Font whose parameters (size, style, typeface) should be represented in the style object
     */
    public DXFStyle(Typeface androidFont)
    {
        // generate a name - just use digits
        this.name = highName++;
        this.androidFont = androidFont;
        this.dxfFontName = androidToDXFFontMap(androidFont);
    }
    
    
    /**
     * Implementation of DXFObject interface method; creates DXF text representing the text style.
     */
    public String toDXFString()
    {
        String returnString = new String();
        
        returnString += "0\nSTYLE\n";
        returnString += "2\n" + name + "\n";
        returnString += "3\n" + dxfFontName + "\n";
        
        // no flags set
        returnString += "70\n0\n";
        
        return returnString;
    }
    
    
    /**
     * Equals method for use in determining if a Style is already present.
     * @param other		Another style representing a font
     * @return			True if the associated Android fonts used to create the styles are equal
     */
    public boolean equals(DXFStyle other)
    {
        if (this.androidFont.equals(other.androidFont))
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    
    
    /**
     * Get the name (integer identifier) for this style object, for use within a DXFText entity.
     * @return
     */
    public String getStyleName()
    {
        String returnString = new String();
        returnString += name;
        return returnString;
    }
    
    
    
    /**
     * Determine a standard DXF font best corresponding to the supplied Android typeface.
     * 
     * Standard DXF type style strings:
     *    STANDARD    
     *    ARIAL   
     *    ARIAL_BOLD
     *    ARIAL_ITALIC
     *    ARIAL_BOLD_ITALIC
     *    ARIAL_BLACK
     *    ISOCPEUR
     *    ISOCPEUR_ITALIC
     *    TIMES
     *    TIMES_BOLD
     *    TIMES_ITALIC
     *    TIMES_BOLD_ITALIC
     *   
     * @param androidFont	Android typeface
     * @return				Corresponding DXF font style string
     */
    private static String androidToDXFFontMap(Typeface androidFont)
    {
    	String returnString;
    	
    	// figure out which typrface to use
    	if (androidFont.equals(Typeface.SERIF))
    	{
        	returnString = "times";
    	}
    	else if (androidFont.equals(Typeface.MONOSPACE))
    	{
    		returnString = "isocpeur";
    	}
    	else
        {
            returnString = "arial";
        }
    	
    	// add bold/italic modifer
    	int fontStyle = androidFont.getStyle();
    	switch (fontStyle)
    	{
    	    case Typeface.BOLD:
    	    {
    	        returnString += "_bold";
    	        break;
    	    }
    	    case Typeface.ITALIC:
            {
                returnString += "_italic";
                break;
            }
    	    case Typeface.BOLD_ITALIC:
            {
                returnString += "_bold_italic";
                break;
            }
            default:
            {
                break;
            }
    	}
    	
    	return returnString;
    }
    
}