/*
 * ADXF Library
 * 
 *   Copyright (C) 2016, Jonathan Sevy <jsevy@cs.drexel.edu>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.adxf; 


import java.io.Serializable;

import android.graphics.Matrix;




/**
 * Utility class representing a 3-dimensional point
 * 
 * @author jsevy
 *
 */
public class RealPoint implements Serializable
{
    private static final long serialVersionUID = 1L;
	
	public float x, y, z;
    
    public RealPoint (float x, float y, float z)
    {
        this.x = x;
        this.y = y;
        this.z = z;
    }
    
    public RealPoint (RealPoint other)
    {
        this.x = other.x;
        this.y = other.y;
        this.z = other.z;
    }
    
   
    public static RealPoint sum(RealPoint p1, RealPoint p2)
    {
        return new RealPoint(p1.x + p2.x, p1.y + p2.y, p1.z + p2.z);
    }
    
    public static RealPoint difference(RealPoint p1, RealPoint p2)
    {
        return new RealPoint(p1.x - p2.x, p1.y - p2.y, p1.z - p2.z);
    }
    
    public static RealPoint scalarProduct(float a, RealPoint p)
    {
        return new RealPoint(a * p.x, a * p.y, a * p.z);
    }
    
    public static RealPoint mapPoint(Matrix m, RealPoint p)
    {
        float[] src = {p.x, p.y};
        float[] dst = new float[2];
        m.mapPoints(dst, src);
        return new RealPoint(dst[0], dst[1], p.z);
    }
    
    public static RealPoint mapVector(Matrix m, RealPoint p)
    {
        float[] src = {p.x, p.y};
        float[] dst = new float[2];
        m.mapVectors(dst, src);
        return new RealPoint(dst[0], dst[1], p.z);
    }
    
    public static float magnitude(RealPoint p)
    {
        return (float)Math.sqrt(p.x*p.x + p.y*p.y + p.z*p.z);
    }
    
    public static float dotProduct(RealPoint p, RealPoint q)
    {
        return p.x*q.x + p.y*q.y + p.z*q.z;
    }
    
    public static RealPoint crossProduct(RealPoint p, RealPoint q)
    {
        RealPoint result = new RealPoint(p.y*q.z-p.z*q.y, p.z*q.x-p.x*q.z, p.x*q.y-p.y*q.x);
        return result;
    }
    
    public static float angleBetween(RealPoint p, RealPoint q)
    {
        float magnitudeProd = magnitude(p)*magnitude(q);
        if (magnitudeProd == 0)
        {
            return 0;
        }
        else
        {
            return (float)Math.acos(dotProduct(p, q)/magnitudeProd);
        }
    }
    
    public boolean equals(Object object)
    {
        if(!(object instanceof RealPoint))
            return false;
            
        RealPoint other = (RealPoint)object;
        
        if ((this.x == other.x) && (this.y == other.y) && (this.z == other.z))
            return true;
        else
            return false;
    }
    
    
}